package asd.td6;

import asd.base.ObjetAClef;

/**
 * Une classe pour les exercices 1, 2, 3, 5 et 6 de la feuille 6.
 */
public class ABR {

    // attribut

    ObjetAClef racine; // la valeur du noeud
    ABR gauche;        // le sous-arbre gauche
    ABR droite;        // le sous-arbre droit

    // constructeurs

    /**
     * Construit un arbre binaire de recherche vide.
     */
    public ABR() {
        gauche = null; // ou droite = null
    }

    /**
     * Construit un arbre binaire de recherche rduit 
     * une feuille de racine x.
     */
    public ABR(ObjetAClef x) {
        racine = x;
        gauche = new ABR();
        droite = new ABR();
    }

    // mthodes primitives

    /**
     * Teste si l'arbre est vide.
     */
    public boolean vide() {
        return gauche == null;
    }

    /**
     * Transforme l'arbre en feuille de valeur x, soit
     * un arbre binaire rduit  une feuille de racine x.
     * Prcondition : this est vide.
     */
    private void feuille(ObjetAClef x) {
        gauche = new ABR();
        droite = new ABR();
        racine = x;
    }

    /**
     * Remplace la racine et les sous-arbres gauche et droit de l'arbre
     * par la racine et les sous-arbres gauche et droit de l'arbre A.
     */
    private void devient(ABR A) {
        racine = A.racine;
        gauche = A.gauche;
        droite = A.droite;
    }

    /**
     * Retire tous les lments de l'arbre.
     */
    private void couper() {
        gauche = null;
        droite = null;
        racine = null;
    }

    // mthodes non primitives

    /**
     * Retourne le nombre d'lments de l'arbre.
     * Complexit : O(n) o n est le nombre d'lments de l'arbre.
     */
    public int taille() {
        if ( vide() )
            return 0;
        return 1 + gauche.taille() + droite.taille();
    }

    // EXO 1 : implmentation de la classe ABR
    // ajout d'un lment (ncessaire pour la classe
    // de test

    /**
     * Ajoute l'objet x  l'arbre. Si un objet de meme clef
     * que x est dj dans l'arbre, ne fait rien.
     * Complexit : O(h) o h est la hauteur de l'arbre.
     */
    public void ajouter(ObjetAClef x) {
        if ( vide() )
            feuille(x);
        else {
            if ( x.clef() < racine.clef() )
                gauche.ajouter(x);
            if ( x.clef() > racine.clef() )
                droite.ajouter(x);
        }
    }


    // EXO 2 : lagage

    /**
     * Retire de l'arbre tous les lments de clef strictement
     * plus petite que k.
     * Complexit : O(h) o h est la hauteur de l'arbre.
     */
    public void elaguerInferieur(int k) {
        if ( ! vide() )
            if ( racine.clef() == k )
                gauche.couper();
            else
                if ( racine.clef() > k )
                    gauche.elaguerInferieur(k);
                else {
                    devient(droite);
                    elaguerInferieur(k);
                }
    }

    /**
     * Retire de l'arbre tous les lments de clef strictement
     * plus grande que k.
     * Complexit : O(h) o h est la hauteur de l'arbre.
     */
    public void elaguerSuperieur(int k) {
        if ( ! vide() )
            if ( racine.clef() == k )
                droite.couper();
            else
                if ( racine.clef() < k )
                    droite.elaguerInferieur(k);
                else {
                    devient(gauche);
                    elaguerSuperieur(k);
                }
    }

    // EXO 3 : conversion ABR <--> tableau tri croissant

    // conversion ABR --> tableau tri croissant
    // Version 1 : utilisation d'un compteur

    /**
     * Retourne un tableau tri croissant contenant exactement
     * les lments de l'arbre.
     * Complexit : O(N) o N est la taille de l'arbre.
     */
    public ObjetAClef[] tableau1() {
        ObjetAClef[] T = new ObjetAClef[taille()];
        Compteur c = new Compteur(0);
        tableau1(T,c);
        return T;
    }

    /**
     * Copie les lments de l'arbre par ordre croissant dans
     * la tranche T[i..N-1] du tableau T avec i = c.valeurCourante().
     * Complexit : O(N-i) o N est la taille de l'arbre.
     */
    private void tableau1(ObjetAClef[] T, Compteur c) {
        if ( ! vide() ) {
            gauche.tableau1(T,c);
            T[c.val] = racine;
            c.val++;
            droite.tableau1(T,c);
        }
    }

    /**
     * Une classe interne pour grer un compteur entier.
     */
    private class Compteur {

        int val; // la valeur du compteur

        /**
         * Construit un compteur de valeur initiale val.
         */
        Compteur(int val) {
            this.val = val;
        }
    }

    // conversion ABR --> tableau tri croissant
    // Version 2 : utilisation d'une fonction

    /**
     * Retourne un tableau tri croissant contenant exactement
     * les lments de l'arbre.
     * Complexit : O(N) o N est la taille de l'arbre.
     */
    public ObjetAClef[] tableau2() {
        ObjetAClef[] T = new ObjetAClef[taille()];
        tableau2(T,0);
        return T;
    }

    /**
     * Copie les lments de l'arbre par ordre croissant dans
     * la tranche T[i..N-1] du tableau T et retourne n le nombre
     * d'lments copis.
     * Complexit : O(N-i) o N est la taille de l'arbre.
     */
    private int tableau2(ObjetAClef[] t, int i) {
        if ( vide() )
            return 0;
        int g = gauche.tableau2(t,i);
        t[i+g] = racine;
        int d = droite.tableau2(t,g+i+1);
        return g+d+1;
    }

    // conversion tableau tri croissant --> ABR

    /**
     * Construit un arbre binaire de recherche contenant
     * exactement les lments du tableau T.
     * Prcondition : le tableau T est croissant et tous ses
     * lments sont de clefs distinctes.
     * Complexit : O(T.length)
     */
    public ABR(ObjetAClef[] T) {
        this();
        remplir(T,0,T.length - 1);
    }

    /**
     * Ajoute dans l'arbre exactement tous les lments de la
     * tranche T[i..j] du tableau T.
     * Prcondition : le tableau T est croissant et tous ses
     * lments sont de clefs distinctes, et l'arbre est vide.
     * Complexit : O(j-i+1) si i <= j, O(1) sinon.
     */
    private void remplir(ObjetAClef[] T, int i, int j) {
        if ( i <= j ) {
            int m = (i + j) / 2;
            feuille(T[m]);
            gauche.remplir(T,i,m-1);
            droite.remplir(T,m+1,j);
        }
    }

    // EXO 5 : Scindage

    /**
     * Rparti les lments de l'arbre de telle sorte que ceux de clef
     * infrieure ou gale  k soient dans G et ceux de clef strictement
     * suprieure  k soient dans D.
     * Prcondition : G et D sont vides.
     */
    public void scinder(int k, ABR G, ABR D) {
        if ( ! vide() )
            if ( racine.clef() <= k ) {
                G.devient(this);
                devient(droite);
                G.droite.couper();
                scinder(k,G.droite,D);
            }
            else {
                D.devient(this);
                devient(gauche);
                D.gauche.couper();
                scinder(k,G,D.gauche);
            }
    }

    // ajout d'un lment

    /**
     * Ajoute l'objet x  la racine de l'arbre.
     */
    public void ajouter2(ObjetAClef x) {
        if ( vide() )
            feuille(x);
        else {
            ABR G = new ABR();
            ABR D = new ABR();
            scinder(x.clef(),G,D);
            racine = x;
            gauche = G;
            droite = D;
        }
    }

    // EXO 6 : Fusion

    /**
     * Fusionne l'arbre A dans l'arbre courant.
     */
    public void fusionnerAvec(ABR A) {
        if ( ! A.vide() ) {
            if ( vide() ) {
                devient(A);
                A.couper();
            }
            else {
                ABR G = new ABR();
                ABR D = new ABR();
                A.scinder(racine.clef(),G,D);
                gauche.fusionnerAvec(G);
                droite.fusionnerAvec(D);
            }
        }
    }

    ///////////////////////////////////////////////////////////////////////

    /**
     * Retourne une chane de longueur k ne contenant que le
     * caractre c.
     */
    private String chaine(char c, int k) {
        String s = "";
        for ( int i = 0; i < k; i++ ) {
            s += c;
        }
        return s;
    }

    /**
     * Affiche l'arbre sous une forme lisible.
     */
    public void afficher() {
        afficher("","");
    }

    /**
     * Affiche l'arbre sous une forme lisible avec r comme prfixe
     * de la chane correspondant  la racine et p comme prfixe
     * des autres chanes.
     */
    private void afficher(String r, String p) {
        if ( vide() ) {
            System.out.println(r);
        }
        else {
            String rs = racine.toString();
            System.out.println(r + rs);
            if ( ! gauche.vide() || ! droite.vide() ) {
                String rr = p + '|' + chaine('_',rs.length()) + ' ';
                droite.afficher(rr, p + '|' + chaine(' ',rs.length() + 1));
                System.out.println(p + '|');
                gauche.afficher(rr, p + chaine(' ',rs.length() + 2));
            }
        }
    }

    ///////////////////////////////////////////////////////////////////////
}
